import click


class DefaultCommandGroup(click.Group):
    """allow a default command for a group"""

    def command(self, *args, **kwargs):
        default_command = kwargs.pop("default_command", False)
        if default_command and not args:
            kwargs["name"] = kwargs.get("name", "<>")
        decorator = super(DefaultCommandGroup, self).command(*args, **kwargs)

        if default_command:

            def new_decorator(f):
                cmd = decorator(f)
                self.default_command = cmd.name
                return cmd

            return new_decorator

        return decorator

    def resolve_command(self, ctx, args):
        try:
            # test if the command parses
            return super(DefaultCommandGroup, self).resolve_command(ctx, args)
        except click.UsageError:
            # command did not parse, assume it is the default command
            args.insert(0, self.default_command)
            return super(DefaultCommandGroup, self).resolve_command(ctx, args)



import subprocess

from versup.conf_reader import get_conf_value


def prepost_script(taskname):
    def doit(function):
        def wrapper(config, version, **kwargs):
            pre_script = get_conf_value(config, "scripts/pre{}".format(taskname))

            if pre_script:
                if kwargs["dryrun"]:
                    print("Execute pre script `{}`\n".format(pre_script))
                else:
                    subprocess.run(pre_script.split() + [version])

            value = function(config, version, **kwargs)

            post_script = get_conf_value(config, "scripts/post{}".format(taskname))
            if post_script:
                if kwargs["dryrun"]:
                    print("Execute post script `{}`\n".format(post_script))
                else:
                    subprocess.run(post_script.split())

            return value

        return wrapper

    return doit



from git import Repo
import os
import semver


def get_repo():
    return Repo(os.getcwd())


def get_username():
    repo = get_repo()
    return repo.config_reader().get_value("user", "name")


def get_email():
    repo = get_repo()
    return repo.config_reader().get_value("user", "email")


def is_repo_dirty():
    repo = get_repo()
    return repo.is_dirty(untracked_files=True)


def get_latest_tag():
    repo = get_repo()
    latest_tags = repo.git.tag(sort="-creatordate").split("\n")[:10]
    for tag in latest_tags:
        try:
            semver.parse_version_info(tag)
        except ValueError:
            continue
        return tag

    raise ValueError


def create_new_tag(new_version, tag_name):
    repo = get_repo()
    repo.create_tag(new_version, message=tag_name)


def create_commit(commit_msg):
    repo = get_repo()
    index = repo.index
    changed_files = [item.a_path for item in repo.index.diff(None)]
    index.add(repo.untracked_files + changed_files)
    repo.git.commit("-m", commit_msg)


def get_commit_messages():
    # git log --pretty=oneline HEAD...0.2.0
    try:
        latest_tag = get_latest_tag()
        commit_range = "HEAD...{}".format(latest_tag)
    except ValueError:
        commit_range = "HEAD"
    repo = get_repo()
    commits = repo.git.log("--pretty=format:%H||%an||%ae||%at||%s", commit_range).split(
        "\n"
    )
    out = []

    for commit in commits:
        data = dict()
        split_commit = commit.split("||")
        if len(split_commit) < 2:
            return out
        data["hash"] = split_commit[0]
        data["author_name"] = split_commit[1]
        data["author_email"] = split_commit[2]
        data["date"] = split_commit[3]
        data["message"] = split_commit[4]

        out.append(data)
    return out



import sys
import os
import click
import semver
import datetime

from versup import __version__
from versup.conf_reader import (
    get_conf_value,
    merge_configs_with_default,
    parse_config_file,
)
from versup.custom_cmd_group import DefaultCommandGroup
import versup.file_updater as file_updater
import versup.gitops as gitops
import versup.template as template
import versup.changelog as changelog
import versup.script_runner as script_runner
from versup.printer import print_ok, print_error, print_warn


class versupContext(object):
    conf = None
    template = None
    version = None


@click.group(cls=DefaultCommandGroup)
@click.pass_context
@click.version_option(version=__version__)
def cli(ctx, **kwargs):
    bobj = versupContext()
    bobj.conf = merge_configs_with_default()
    bobj.template_data = template.token_data
    ctx.obj = bobj


@cli.command()
@click.pass_context
@click.option("-l", "--local", is_flag=True)
@click.option("-g", "--global", is_flag=True)
def show_config(ctx, **kwargs):
    import pprint

    config = ctx.obj.conf
    if kwargs["local"]:
        config = parse_config_file("./.versup.json")
    if kwargs["global"]:
        config = parse_config_file("~/.config/versup.json")
    pprint.pprint(config)


@cli.command(default_command=True, name="increment/version")
@click.pass_context
@click.argument("increment")
@click.option("--no-commit", is_flag=True, help="Skip making commit")
@click.option("--no-changelog", is_flag=True, help="Skip changelog update")
@click.option("--no-tag", is_flag=True, help="Skip creating tag")
@click.option("-n", "--dryrun", is_flag=True, help="Show what will be done.")
def do_versup(ctx, **kwargs):
    """
    Increment or set project version
    """
    ctx.obj.version = kwargs["increment"]

    if not ctx.obj.version in get_conf_value(ctx.obj.conf, "version/increments"):
        try:
            # Parse to see if version format is ok
            semver.parse_version_info(ctx.obj.version)
        except ValueError:
            print_error("Supplied version is invalid")
            return

    version = get_new_version(ctx.obj.conf, ctx.obj.version, **kwargs)

    # Update the token_data with what we know
    template.token_data["version"] = version

    today = datetime.date.today()
    template_date = today.strftime(get_conf_value(ctx.obj.conf, "tokens/date/format"))
    template.token_data["date"] = template_date

    template_date = today.strftime(
        get_conf_value(ctx.obj.conf, "tokens/version_date/format")
    )
    template.token_data["version_date"] = template_date

    template.token_data["author_name"] = gitops.get_username()
    template.token_data["author_email"] = gitops.get_email()

    template.token_data["message"] = get_conf_value(ctx.obj.conf, "commit/message")

    apply_bump(ctx.obj.conf, version, **kwargs)


@script_runner.prepost_script("bump")
def apply_bump(config, version, **kwargs):
    # Run through all stages of a release
    # Update the files specified in config
    files_to_update = get_conf_value(config, "files")
    file_updater.update_files(version, files_to_update, kwargs["dryrun"])

    # create changelog
    if not kwargs["no_changelog"] and get_conf_value(config, "changelog/enabled"):
        do_changelog(config, version, **kwargs)

    # create new commit with version
    if not kwargs["no_commit"] and get_conf_value(config, "commit/enabled"):
        commit(config, version, **kwargs)

    # tag commit (only if a commit is made)
    if (
        not kwargs["no_commit"]
        and not kwargs["no_tag"]
        and get_conf_value(config, "tag/enabled")
    ):
        tag(config, version, **kwargs)


def bump_version(latest_version, increment):
    func = "bump_{}".format(increment)
    return semver.__dict__[func](latest_version)


def get_new_version(config, version, **kwargs):
    """
    version is either an increment or a semantic version. Given an increment
    the current version (based on the latest git commit, or the initial version
    from the config) is incremented.
    Given a version, that version is used as is provided it is valid
    """
    if gitops.is_repo_dirty():
        if kwargs["dryrun"]:
            print_warn("Repo is dirty.")
        else:
            print_error("Repo is dirty. Cannot continue")
            # TODO raise exception
            sys.exit(1)

    try:
        latest_version = gitops.get_latest_tag()
    except:
        latest_version = get_conf_value(config, "version/initial")

    # TODO: What happens if there's no latest version?
    if version in get_conf_value(config, "version/increments"):
        new_version = bump_version(latest_version, version)
    else:
        try:
            semver.parse_version_info(version)
            new_version = version
        except ValueError:
            print_error("Supplied version is not a valid SemVer string or increment")
            sys.exit(1)

    # Update value in template data struct
    template.token_data["version"] = new_version

    return new_version


@script_runner.prepost_script("changelog")
def do_changelog(config, version, **kwargs):
    if not get_conf_value(config, "changelog/enabled"):
        return
    changelog_config = get_conf_value(config, "changelog")
    changelog_file = changelog_config["file"]
    # If no changelog file and create is off, prompt
    if not kwargs["dryrun"] and not os.path.isfile(changelog_file):
        if not changelog_config["create"]:
            if not click.confirm("No changelog file found. Create it?"):
                return
            # Ok to create/update it now
    changelog.write(
        changelog_file,
        changelog_config["version"],
        changelog_config["commit"],
        changelog_config["separator"],
        changelog_config["open"],
        version,
        kwargs["dryrun"],
    )
    print_ok("Changelog updated")


@script_runner.prepost_script("commit")
def commit(config, version, **kwargs):
    if not get_conf_value(config, "commit/enabled"):
        return

    if not kwargs["dryrun"] and not gitops.is_repo_dirty():
        print("No unstaged changes to repo. Making no new commit.")
        return
    commit_config = get_conf_value(config, "commit")
    template.token_data["version"] = version
    commit_msg = template.render(commit_config["message"])
    if kwargs["dryrun"]:
        print("Create commit with commit msg: {}".format(commit_msg))
    else:
        gitops.create_commit(commit_msg)
    print_ok("Commit created")


@script_runner.prepost_script("tag")
def tag(config, version, **kwargs):
    if not get_conf_value(config, "tag/enabled"):
        return

    if not kwargs["dryrun"] and gitops.is_repo_dirty():
        print("Unstaged changes to repo. Cannot make a tag.")
        sys.exit(1)
    tag_config = get_conf_value(config, "tag")
    template.token_data["version"] = version
    tag_name = template.render(tag_config["name"])
    if kwargs["dryrun"]:
        print("Create tag {} with msg {}".format(version, tag_name))
    else:
        gitops.create_new_tag(version, tag_name)

    print_ok("Tag {} created".format(tag_name))


def main():
    cli()



from versup.conf_reader import get_conf_value
import versup.gitops as gitops
import sys
import versup.template as template
from subprocess import run, PIPE
import sys


def show_file(changelog_file):
    """
    Open the supplied file with $EDITOR

    :changelog_file: The path to the file to open.

    """
    with open(changelog_file, "r") as fh:
        data = fh.read()

    try:
        pager = run(
            ["less", "-F", "-R", "-S", "-X", "-K"],
            stdout=sys.stdout,
            input=data,
            encoding="ascii",
        )
    except KeyboardInterrupt:
        # let less handle this, -K will exit cleanly
        pass


def write(
    changelog_file, version_line, changelog_line, separator, show, version, dryrun=False
):
    """
    Write the new changelog file. Gets the changelog filename (and optional
    path) from the config. Parses the commit messages and prepends the
    commits to the original text and saves out the file.
    Creates the file if there it doesn't exist

    :version: The new version to use
    :dryrun: Whether this is a dryrun. If true does not write file but
             prints new content to stdout

    """

    commits = gitops.get_commit_messages()

    if dryrun:
        print("Writing changelog entries:\n")
        for commit_data in commits:
            commit_data["hash"] = commit_data["hash"]
            commit_data["hash4"] = commit_data["hash"][:4]
            commit_data["hash7"] = commit_data["hash"][:7]
            commit_data["hash8"] = commit_data["hash"][:8]

            commit_line = template.render(changelog_line, commit_data)
            print(commit_line)
        print(separator)
    else:
        # Read original changelog
        try:
            with open(changelog_file, "r") as fh:
                original_data = fh.read()
        except FileNotFoundError:
            original_data = ""

        version = template.render(version_line, {"version": version})
        with open(changelog_file, "w+") as fh:
            fh.write(version + "\n")
            for commit_data in commits:
                commit_data["hash"] = commit_data["hash"]
                commit_data["hash4"] = commit_data["hash"][:4]
                commit_data["hash7"] = commit_data["hash"][:7]
                commit_data["hash8"] = commit_data["hash"][:8]

                commit_line = template.render(changelog_line, commit_data)
                fh.write(commit_line + "\n")

            fh.write(separator + original_data)

        if show:
            show_file(changelog_file)



import os
import json
from versup.default_conf import default_conf


config_files = [
    "~/.config/versup.json",
    "./.versup.json",
]


def write_default_to_home():
    home_file = os.path.expanduser(config_files[0])
    if not os.path.isfile(home_file):
        with open(home_file, "w+") as f:
            f.write(json.dumps(default_conf, indent=4))


def parse_config_file(config_file):
    cwd = os.getcwd()
    d = {}
    try:
        with open(os.path.expanduser(config_file), "r") as conf_file:
            d = json.loads(conf_file.read())
    except FileNotFoundError:
        pass
    return d


def merge(a, b, path=None):
    "merges b into a"
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def merge_configs_with_default():
    write_default_to_home()
    current_config = {}
    for config_file in config_files:
        config = parse_config_file(config_file)
        merge(current_config, config)
    return current_config


def get_conf_value(config_data, key_path):
    paths = key_path.split("/")
    root = config_data

    for p in paths:
        try:
            v = root[p]
            root = v
        except KeyError:
            # TODO: Raise here?
            return None
    return v



default_conf = {
    "force": False,  # Force the command without prompting the user
    "silent": False,  # Minimize the amount of logs
    "files": {},  # A map of `relativeFilePath: [regex, replacement, regexFlags?] | [regex, replacement, regexFlags?][]`
    "version": {
        "enabled": True,  # Bump the version number
        "initial": "0.0.0",  # Initial version
        "increments": [
            "major",
            "minor",
            "patch",
            "premajor",
            "preminor",
            "prepatch",
            "prerelease",
            "custom",
        ],  # List of available increments to pick from
    },
    "changelog": {
        "enabled": True,  # Enable changelog auto-updates
        "create": False,  # Create the changelog file if it doesn"t exist
        "open": True,  # Open the changelog file after bumping
        "file": "CHANGELOG.md",  # Name of the changelog file
        "version": "### Version [version]",  # Template for the version line
        "commit": "- [message]",  # Template for the commit line
        "separator": "\n",  # Template for the separator between versions sections
    },
    "commit": {
        "enabled": True,  # Commit the changes automatically
        "message": "Update version to [version]",  # Template for the commit message
    },
    "tag": {
        "enabled": True,  # Tag the bump commit
        "name": "v[version]",  # Template for the name of the tag
    },
    "tokens": {
        "date": {
            "format": "%Y-%m-%d"  # Python datetime format to use when generating the `[date]` token
        },
        "version_date": {
            "format": "%Y-%m-%d"  # Python datetime format to use when generating the `[version_date]` token
        },
    },
    "scripts": {
        "prebump": "",  # Script to execute before bumping the version
        "postbump": "",  # Script to execute after bumping the version
        "prechangelog": "",  # Script to execute before updating the changelog
        "postchangelog": "",  # Script to execute after updating the changelog
        "precommit": "",  # Script to execute before committing
        "postcommit": "",  # Script to execute after committing
        "pretag": "",  # Script to execute before tagging
        "posttag": ""  # Script to execute after tagging
    },
}



from colorama import Fore, Back, Style


def print_ok(message):
    print(Fore.GREEN + u"\u2713 " + Fore.RESET + message + "\n")


def print_error(message):
    print(Fore.RED + Style.BRIGHT + u"\u2717 " + Style.RESET_ALL + message + "\n")


def print_warn(message):
    print(Fore.RED + Style.BRIGHT + u"! " + Style.RESET_ALL + message + "\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['versup']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'colorama>=0.4.3,<0.5.0',
 'gitpython>=3.0,<4.0',
 'semver>=2.8,<3.0']

entry_points = \
{'console_scripts': ['versup = versup.__main__:main']}

setup_kwargs = {
    'name': 'versup',
    'version': '0.3.2',
    'description': '',
    'long_description': '# versup\n\nMIT license\n\n[![Documentation Status](https://readthedocs.org/projects/versup/badge/?version=latest)](https://versup.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/versup.svg)](https://badge.fury.io/py/versup)\n\n** Still in early stages, so use with caution **\n\nBump your project version, update version numbers in your files, create a changelog,\nmake a commit, and tag it, all in one easy step. versup is also quite configurable.\n\n# Install\n\nInstall with either poetry\n\n`poetry install`\n\nor pip\n\n`pip install .`\n\n# Quick start\n\nTo get started all versup needs to know is the new version increment or number.\nYou can provide it with a valid semantic version increase such as `patch`, `minor`,\n`major` etc, or an entirely new semantic version like `1.2.5`.\n\nIf you specifiy a version number, then versup will take that version and apply\nit to the current project as is. If you provide an increment, it will get the\nlast version number from either the latest git tag that has a valid version,\nor from the default version in the config file.\n\n# Configuration\n\nOne intial launch, versup copies a default config to your home directory (`~/.config/versup.json`) which has some good defaults.\n\n```\n{\n    "force": False,  # Force the command without prompting the user\n    "silent": False,  # Minimize the amount of logs\n    "files": {},  # A map of `relativeFilePath: [regex, replacement, regexFlags?] | [regex, replacement, regexFlags?][]`\n    "version": {\n        "enabled": True,  # Bump the version number\n        "initial": "0.0.0",  # Initial version\n        "increments": [\n            "major",\n            "minor",\n            "patch",\n            "premajor",\n            "preminor",\n            "prepatch",\n            "prerelease",\n            "custom",\n        ],  # List of available increments to pick from\n    },\n    "changelog": {\n        "enabled": True,  # Enable changelog auto-updates\n        "create": False,  # Create the changelog file if it doesn"t exist\n        "open": True,  # Open the changelog file after bumping\n        "file": "CHANGELOG.md",  # Name of the changelog file\n        "version": "### Version [version]",  # Template for the version line\n        "commit": "- [message]",  # Template for the commit line\n        "separator": "\\n",  # Template for the separator between versions sections\n    },\n    "commit": {\n        "enabled": True,  # Commit the changes automatically\n        "message": "Update version to [version]",  # Template for the commit message\n    },\n    "tag": {\n        "enabled": True,  # Tag the bump commit\n        "name": "v[version]",  # Template for the name of the tag\n    },\n    "tokens": {\n        "date": {\n            "format": "%Y-%m-%d"  # Python datetime format to use when generating the `[date]` token\n        },\n        "version_date": {\n            "format": "%Y-%m-%d"  # Python datetime format to use when generating the `[version_date]` token\n        },\n    },\n    "scripts": {\n        "prebump": "",  # Script to execute before bumping the version\n        "postbump": "",  # Script to execute after bumping the version\n        "prechangelog": "",  # Script to execute before updating the changelog\n        "postchangelog": "",  # Script to execute after updating the changelog\n        "precommit": "",  # Script to execute before committing\n        "postcommit": "",  # Script to execute after committing\n        "pretag": "",  # Script to execute before tagging\n        "posttag": "",  # Script to execute after tagging\n        "prerelease": "",  # Script to execute before releasing\n        "postrelease": "",  # Script to execute after releasing\n    },\n}\n```\n\nYou can edit this file to affect all bumps, or create a `.versup.json` file in your project root\nand versup will use these values to override the global ones.\n\n# Template tags\n\nIn various places you can define what text to use for commit messages, or tags etc.\nThese support tag fields that are replaced with information. Know fields are:\n\n- version: The new version\n- message: The new commit message\n- date: Today\'s date formatted according to `tokens/date/format` in the config\n- hash: The new commit hash, full length\n- hash4: The new commit hash, first four characters\n- hash7: The new commit hash, first seven characters\n- hash8: The new commit hash, first eight characters\n- author_name: The author name from the git config\n- author_email: The author email from the git config\n\n# Updating files\n\nversup can update versions in files. The way this works is by configuring a regex\nfor each file that you want to update. So for example:\n\n```\n"files": {\n    "README.rst": [\n      ["Version ([\\\\d\\\\.]+) ", "Version [version] "],\n      ["Version is ([\\\\d\\\\.]+)", "Version is [version]"]\n    ]\n  },\n```\n\nHere the file `README.rst` is updated by matching a regex `Version ([\\\\d\\\\.]+)`\nwhich will match any text like `Version 1.3` or `Version 1.3.7`. They are standard\nregular expressions. The text that is matched is then replaced with the next argument\n`Version [version]` where `[version]` is the new version. You can regex and replace on\nanything really.\n\n# Scripts\n\nThere are a number of pre and post scripts that can be executed at various\nstages of the bump process. These are under the `scripts` section. They are\ncalled as-is and receive the new version number as the only argument. They\ncan be anything, shell scripts, python scripts, etc, but they must be\nexecutable in a regular shell, as they will be invoked as such.\n',
    'author': 'Sven Steinbauer',
    'author_email': 'sven@unlogic.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/unlogic/versup',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import re
import versup.template as template
from versup.conf_reader import get_conf_value
from colorama import Style


def update_file_data(data, replace_list):
    regex = replace_list[0]
    new_text = replace_list[1]

    updated_data = re.sub(regex, new_text, data, flags=re.M)
    return updated_data


def show_updates(filename, data, replace_list):
    regex = replace_list[0]
    new_text = replace_list[1]

    lines = data.split("\n")
    for line in lines:
        line = line.strip()
        m = re.match(regex, line)
        if m:
            updated_data = re.sub(regex, new_text, line, flags=re.M)
            print(
                "In file {3}{0}{4} replace {3}{1}{4} with {3}{2}{4}".format(
                    filename, line, updated_data, Style.BRIGHT, Style.RESET_ALL
                )
            )


def update_files(new_version, files, dryrun):
    if not files:
        # No files to update
        return

    filenames = list(files.keys())
    template_data = {"version": new_version}

    for filename in filenames:
        try:
            with open(filename, "r") as file_h:
                data = file_h.read()
        except FileNotFoundError:
            print("Unable to find {} to update.".format(filename))
            continue

        # If it's only one entry make it a list to simplify program flow
        if not any(isinstance(el, list) for el in files[filename]):
            files[filename] = [files[filename]]

        for replace in files[filename]:
            replace[1] = template.render(replace[1], template_data)
            if dryrun:
                show_updates(filename, data, replace)
            else:
                data = update_file_data(data, replace)

        if not dryrun:
            with open(filename, "w") as file_h:
                file_h.write(data)



import re


token_data = {
    "version": None,
    "version_date": None,
    "message": None,
    "date": None,
    "hash": None,
    "hash4": None,
    "hash7": None,
    "hash8": None,
    "author_name": None,
    "author_email": None,
}


def render(string, data={}):
    # Merge new data with default
    new_data = {**token_data, **data}
    for k, v in new_data.items():
        # Check if value is not None and if the key is valid
        if v and (k in token_data.keys()):
            string = re.sub(r"\[{}\]".format(k), v, string, flags=re.M)

    return string
